# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

'''
Image data packets - list of packets containing image header and data chunks.

Image header must contain the following information:
    1. Timestamp
    2. Number of chunks, or -1/0 if live data stream
    3. Feed name
'''

MAX_PAYLOAD_SIZE = 60000
JPEG_MAX_COMPRESSION_RATIO = 80 # percent


import io
import json
import struct
import asyncio
import traceback
import numpy as np
from PIL import Image
from datetime import datetime

from .packet import Packet, PacketType
from .data import DataPacket


# Image header packet
class ImageFeedbackPacket(DataPacket):
    pType = PacketType.IMAGE_FDBK

    # Empty Constructor
    def __init__(self):
        super().__init__()


# Image start chunk
class ImageStartPacket(Packet):
    pType = PacketType.IMAGE_START

    # Empty Constructor
    def __init__(self):
        super().__init__()
    

    # Encode chunk in packet
    def encode(self, numChunks, firstChunk, feedName='PRIMARY'):

        # First 8 bytes represent UTC UNIX timestamp as 64-bit float (double-precision)
        utc = datetime.utcnow().timestamp()

        # Construct feedname
        feedNameBytes = bytes(''.join([ feedName[i] if i < len(feedName) else chr(0) for i in range(12) ]), 'utf-8')

        self.pData = bytes(struct.pack("d", utc)) + feedNameBytes + bytes([int(numChunks)]) + firstChunk


    # Get timestamp from packet
    def getTimestamp(self):
        utc = struct.unpack("d", self.pData[:8])[0]
        return datetime.fromtimestamp(utc)

    # Get image/video feed name
    def getFeedName(self):
        return ''.join([ chr(c) for c in self.pData[8:20] if c > 0 ])

    # Get number of chunks from packet
    def getNumChunks(self):
        return self.pData[20]

    # Get chunk from packet
    def getChunk(self):
        return self.pData[21:]


# Image chunk packet
class ImageChunkPacket(Packet):
    pType = PacketType.IMAGE_CHUNK

    # Empty Constructor
    def __init__(self):
        super().__init__()


class ImageData:
    '''
    We can't transmit an image with only a single packet, so this is a set of packets
    '''

    # First chunk is the start packet
    chunks = list() # List of chunks
    

    # Constructor
    def __init__(self, chunks):
        self.chunks = chunks


    @staticmethod
    def constructFromBytes(
        imageData,
        timestamp=None,
        feed="PRIMARY"
    ):
        '''
        Generate an image packet set given a pre-compressed image binary, and optional parameters.

        Parameters
        ----------
        @param imageBytes [bytes] - image bytes data
        @param timestamp [datetime] - if None, use current timestamp, and set to Live
        @param feed [str] - feed name
        '''

        imageChunks = list()

        # Construct image data packets (chunkify)
        for chunkStart in range(0,len(imageData),MAX_PAYLOAD_SIZE):

            chunkEnd = chunkStart + MAX_PAYLOAD_SIZE

            if chunkStart == 0:
                chunkPacket = ImageStartPacket()
                chunkPacket.encode(np.ceil(len(imageData) / MAX_PAYLOAD_SIZE), imageData[chunkStart:chunkEnd], feedName=feed)

            else:
                chunkPacket = ImageChunkPacket()
                chunkPacket.pData = imageData[chunkStart:chunkEnd]
            
            imageChunks.append(chunkPacket)

        return ImageData(imageChunks)


    @staticmethod
    def constructFromImage(
            image,
            timestamp=None,
            feed="PRIMARY",
            resize=None,
            compressionRatio=JPEG_MAX_COMPRESSION_RATIO
        ):
        '''
        Generate an image packet set given an image, and optional parameters.

        Parameters
        ----------
        @param image [np.ndarray] - image data
        @param timestamp [datetime] - if None, use current timestamp, and set to Live
        @param feed [str] - feed name
        @param resize - None (don't resize), or (w,h)
        @param compressionRatio - percentage JPEG compression

        Returns
        -------
        @out [ImageData]
        '''
        
        # Construct image data and then call the bytes constructor
        imageData = io.BytesIO()
        imageJPEG = Image.fromarray(image.astype(np.uint8))

        if resize is not None:
            imageJPEG = imageJPEG.resize(resize)

        # Compress JPEG for transmission (sorry, but we need fast transmissions)
        imageJPEG.save(imageData, format='JPEG', quality=compressionRatio)

        return ImageData.constructFromBytes(
            imageData.getvalue(),
            timestamp=timestamp,
            feed=feed
        )

    
    def getTimestamp(self):
        '''
        Return timestamp from when the image was created/sent.

        Returns
        -------
        @out timestamp [datetime] - image timestamp
        '''

        if len(self.chunks) < 1:
            raise Exception('Image does not have any data')

        return self.chunks[0].getTimestamp()


    def getImage(self):
        '''
        Return image as numpy array from packet data

        Returns
        -------
        @out image [np.ndarray] - image as numpy array
        '''

        if len(self.chunks) < 1:
            raise Exception('Image does not have any data')
        
        # Construct image as byte array
        imageData = bytes()

        for chunkId in range(len(self.chunks)):
            
            if chunkId == 0:
                chunk = self.chunks[chunkId].getChunk()

            else:
                chunk = self.chunks[chunkId].pData
            
            imageData += chunk


        imageJPEG = Image.open(io.BytesIO(imageData))
        
        # Generate numpy array from byte array
        image = np.array(imageJPEG, dtype=np.uint8)

        return image

    
    # Transmit class function
    async def transmit(self, socket):
        '''
        Transmit packet via socket.

        Parameters
        ----------
        @param socket [socket.socket] - connected socket object/reference

        Returns
        -------
        @out [bool] - true if success
        '''

        try:

            # Transmit image data chunks (including first)
            for chunk in self.chunks:
                await chunk.transmit(socket)

            return True # Success

        except Exception as e:
            print(e)
            print(traceback.format_exc())
            pass

    
    # Parse static method (parse set of packets)
    @staticmethod
    async def parse(imageHeader, socket):
        '''
        Parse image data packets.

        Parameters
        ----------
        @param imageHeader [ImageStartPacket] - first chunk
        @param socket [socket.socket] - connected socket object/reference to receive the image data chunks

        Returns
        -------
        @out [ImageData] - parsed image data packet set
        '''

        # Extract image data - header and chunks
        imageChunks = [ imageHeader ]

        numChunks = imageHeader.getNumChunks()

        # Parse chunks
        for _ in range(1,numChunks):
            chunkPayload = await socket.recv()

            # Ensure chunk is actually a chunk
            if PacketType(bytes([chunkPayload[0]])) != PacketType.IMAGE_CHUNK:
                raise Exception('Unexpected packet type received as image chunk.')

            imageChunk = ImageChunkPacket()
            imageChunk.pData = chunkPayload[1:]
            imageChunks.append(imageChunk)
            
        return ImageData(imageChunks)


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

import asyncio

from .protocol import *

# socket handler
async def connectionHandler(node):
    '''
    socket handling task - to be started as a thread.

    Parameters
    ----------
    @param node [NodeLink] - NodeLink object
    '''
    
    while node.threadsActive: # Keep socket open

        # Wait for packet
        p = await Packet.receive(node.socket)

        # Node can only receive image feedback, manual command and mission update packets
        if p.pType == PacketType.IMAGE_FDBK:

            p.__class__ = ImageFeedbackPacket
            feedbackData = p.getData()

            node.imageLatency = feedbackData["LATENCY"]
            node.imageFrameRate = feedbackData["FRAME_RATE"]

            print(feedbackData)

            if node.imageQuality > 0 and (feedbackData["LATENCY"] > 0.05 or feedbackData["FRAME_RATE"] < 15):
                node.imageQuality -= 1 # Drop quality

            elif node.imageQuality < 3 and (feedbackData["LATENCY"] < 0.01 and feedbackData["FRAME_RATE"] > 15):
                node.imageQuality += 1 # Increase quality

            # print(node.imageQuality)
            # Else, maintain quality


        elif p.pType == PacketType.MANUAL_CMD:
            
            p.__class__ = ControlPacket

            # Manual control callback
            if node.manualControlCallback is not None:
                await node.manualControlCallback(p.getData()) # ControlPacket.getData returns a dictionary


        elif p.pType == PacketType.MISSION:

            p.__class__ = MissionPacket

            # Mission update callback
            if node.missionUpdateCallback is not None:

                mission = p.getData()

                # If FETCH is set, then re-fetch mission
                if "FETCH" in mission and mission["FETCH"]:
                    mission = await node.fetchMission() # Re-fetch mission

                await node.missionUpdateCallback(mission) # MissionPacket.getData returns a dictionary


        else:

            print("Received unexpected packet:", p.pType)
            continue


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

DATA_BUFFER_SIZE = 65536 #64kB

import os
import time
import json
import errno
import socket
import asyncio
import requests
import threading
import websockets
import subprocess

from .protocol import *

from .connection import connectionHandler

IMAGE_QUALITY_SIZE = [(360,240), (480,360), (640,480), (960,720)]

# Get IP location
ipLoc = [ float(i) for i in requests.get('http://ipinfo.io/loc').text[:-1].split(',') ]

# NodeLink class definition
class NodeLink:
    '''
    A NodeLink object is used to interface with the Imperix Cloud.
    See the README.md file or read the complete documentation at:
    https://docs.imperix.ai
    '''

    socket = None

    connectionRetries = 0

    # Video streams
    videoStreams = {} # Feed name to process dict

    # Image transmission feedback for auto-optimization
    imageQuality = 3 # Best quality
    imageLatency = 0
    imageFrameRate = 0


    # Constructor
    def __init__(self, missionUpdateCallback=None, manualControlCallback=None):
        '''
        Initialize link to Imperix Cloud for the node with environment config, and provided mission update callback and manual control callback.

        Parameters
        ----------
        @param missionUpdateCallback [func(mission)] - callback function to be called when node's mission is updated.
        @param manualControlCallback [func(control)] - callback function to be called when a manual control command is received.
        '''

        self.threadsActive = True
        self.videoStreams = {}

        # Set manual control callback thread
        self.manualControlCallback = manualControlCallback
        self.missionUpdateCallback = missionUpdateCallback


    # Destructor
    def __del__(self):
        self.disconnect() # Disconnect before deletion to kill threads


    # Connect
    async def connect(self, config='node.cfg', reconnect=False):

        try:

            # Load connection config
            with open(config, 'r') as cfgFile:
                cfg = json.load(cfgFile)

            # Create socket to streamer
            self.socket = await websockets.connect(cfg["STREAMER_URI"])

            # Send authorization packet
            p = AuthPacket.constructFromAuth(
                cfg["NODE_UUID"],
                cfg["ACCESS_KEY"],
                isNode=True
            )
            
            await p.transmit(self.socket) # Transmit auth packet

            print("Connected to Imperix Cloud")

        except Exception as e:
            print("Failed to connect to Imperix Cloud")
            print(e)
            return False


        if not reconnect:
            asyncio.ensure_future(self.start()) # Start processing connection handler coroutine

        


    # Coroutine
    async def start(self):

        while True: # Try again if we loose connection

            if self.socket is None:
                await asyncio.sleep(1) # Wait to connect

            try:
                await connectionHandler(self)
                self.connectionRetries = 0

            except Exception as e:
                print("NodeLink Socket Error:", e)
                self.connectionRetries += 1
                await asyncio.sleep(1) # Wait and try again

                # If we exceed 3 retries, re-connect
                if self.connectionRetries >= 3:
                    print("Attempting to re-connect to socket")
                    await self.connect(reconnect=True)


    # Get mission
    async def fetchMission(self):
        
        pass


    # Mission update
    async def updateMission(self, activeWaypoint, missionStatus):
        '''
        Update mission state for display on Imperix Commander.

        Parameters
        ----------
        @param activeWaypoint [int] - active waypoint index, must be less than number of waypoints
        @param missionStatus [str] - mission status/error
        '''
        
        packet = MissionPacket()
        packet.setMissionStatus(
            activeWaypoint=activeWaypoint,
            missionStatus=missionStatus
        )

        await packet.transmit(self.socket)


    # Streaming data to the cloud
    # Telemetry
    async def transmitTelemetry(self, telemetry):
        '''
        Transmit dict with JSON-serializable telemetry parameters to Imperix Cloud.

        Parameters
        ----------
        @param telemetry [dict] - see README.md or documentation.
        '''

        # Ensure lat/lon exists
        if "LOC_LATITUDE" not in telemetry or "LOC_LONGITUDE" not in telemetry or telemetry["LOC_LATITUDE"] == None or telemetry["LOC_LONGITUDE"] == None or telemetry["LOC_LATITUDE"] == 0 or telemetry["LOC_LONGITUDE"] == 0:
            telemetry["LOC_LATITUDE"] = ipLoc[0]
            telemetry["LOC_LONGITUDE"] = ipLoc[1]
            telemetry["LOC_SOURCE"] = "IP"

        packet = TelemetryPacket()

        packet.setData(telemetry)
        await packet.transmit(self.socket)

    
    # Image (array)
    async def transmitImage(self, image, timeStamp=None, feed='PRIMARY', optimize=True):
        '''
        Transmit image as numpy array to Imperix Cloud.

        Parameters
        ----------
        @param image [np.ndarray] - numpy array image (1 or 3 channels)
        @param timeStamp [datetime.datetime] - timestamp as a datetime object
        @param feed [str] - name of image/video feed
        @param optimize [bool] - auto-optimize framerate?
        '''

        # Set image transmission parameters to optimize for frame-rate
        if optimize:
            p = ImageData.constructFromImage(image, timeStamp, feed, resize=IMAGE_QUALITY_SIZE[self.imageQuality], compressionRatio=70)

        else:
            p = ImageData.constructFromImage(image, timeStamp, feed)
        
        await p.transmit(self.socket)

    
    # Image (binary)
    async def transmitImageBinary(self, image, timeStamp=None, feed='PRIMARY'):
        '''
        Transmit image as numpy array to Imperix Cloud.

        Parameters
        ----------
        @param image [bytes] - JPEG images as bytes object
        @param timeStamp [datetime.datetime] - timestamp as a datetime object
        @param feed [str] - name of image/video feed
        '''

        p = ImageData.constructFromBytes(image, timeStamp, feed)
        await p.transmit(self.socket)


    # Encoded video stream (MPEG)
    async def streamVideoFromSource(self, feedName, source, frameRate=30, videoSize=(640,480), isFile=False):
        '''
        Stream live video from source - file or camera.
        Stream ends (and encoding subprocess is killed) at end of file, or if camera stream is empty.

        Parameters
        ----------
        @param feedName [str] - video feed name
        @param source [str] - video source
        @param frameRate [int] - video frame-rate
        @param videoSize [tuple.int] - video size (w,h)
        @param isFile [bool] - video source is a file?
        '''

        # Start encoding subprocess, add to map
        udpPort = 9000

        self.videoStreams[feedName] = {}

        # Start UDP socket
        self.videoStreams[feedName]["UDP"] = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        # Try to bind on port until success
        for _ in range(10): # Max 10 attempts/streams
            try:
                self.videoStreams[feedName]["UDP"].bind(("localhost", udpPort))
                break # Success
                
            except:
                udpPort += 1 # Try next port
        

        # Start UDP server to handle FFMPEG stream
        self.videoStreams[feedName]["KILL"] = False

        # Threading
        # self.videoStreams[feedName]["THREAD"] = threading.Thread(target=udpPacketHandler, args=(self, feedName))
        # self.videoStreams[feedName]["THREAD"].start()

        # Asyncio coroutine
        asyncio.ensure_future(self.udpServer(feedName))

        # Start FFMPEG streaming
        self.videoStreams[feedName]["PROC"] = subprocess.Popen([
            "ffmpeg",

            # Input
            "-f", "v4l2",
            "-framerate", str(frameRate),
            "-video_size", f"{videoSize[0]}x{videoSize[1]}",
            "-i", source,

            # Settings
            "-preset", "ultrafast",
            "-tune", "zerolatency",
            "-threads", str(1),

            # Output
            "-vcodec", "mpeg1video",
            "-b", "900k", # Bit rate
            "-f", "mpegts",
            f"udp://localhost:{udpPort}" 
        ])

        return

    
    # Video stream UDP server packet handler
    async def udpServer(self, feedName):

        sock = self.videoStreams[feedName]["UDP"]
        sock.setblocking(0) # Non-blocking call

        # Run until either all node threads are killed, or specific feed is killed
        while self.threadsActive and not self.videoStreams[feedName]["KILL"]:

            try:
                rx,_ = sock.recvfrom(DATA_BUFFER_SIZE)
            
            except socket.error as e:
                if e.errno != errno.EAGAIN:
                    raise e

                await asyncio.sleep(0.01)
                continue

            except Exception as e:
                print(e)
                # TODO - handle broken connection
                break

            p = VideoPacket.constructFromChunk(rx, feedName=feedName)
            await p.transmit(self.socket)
        

    # Data stream
    async def transmitData(self, data):
        '''
        Transmit any JSON-serializable data to Imperix Cloud.

        Parameters
        ----------
        @param data [dict] - JSON-serializable data dictionary
        '''

        packet = DataPacket()
        packet.setData(data)
        await packet.transmit(self.socket)

    
    # Disconnect and close threads
    async def disconnect(self):
        '''
        Disconnect and close sockets, and kill threads.
        '''

        # Kill connection handler thread
        self.threadsActive = False

        # Kill video encoding subprocesses
        for feed in self.videoStreams:
            self.videoStreams[feed]["PROC"].terminate()
        #     self.videoStreams[feed]["KILL"] = True
        #     self.videoStreams[feed]["THREAD"].join()

        # Close socket
        self.socket.close()


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

from .packet import Packet, PacketType
from .auth import AuthPacket
from .data import DataPacket
from .telemetry import TelemetryPacket
from .control import ControlPacket
from .image import ImageData, ImageFeedbackPacket, ImageStartPacket, ImageChunkPacket
from .video import VideoPacket
from .request import RequestPacket
from .mission import MissionPacket


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

'''
Mission update data packet.
Inherited from DataPacket class, implements:
    1. Constructor given mission data as a dict
    2. Parser to get mission data as a dict
'''

from .packet import PacketType
from .data import DataPacket

class MissionPacket(DataPacket):
    '''
    Mission update data packet object. (see above)
    '''

    pType = PacketType.MISSION
    pData = bytes()

    # Empty Constructor
    def __init__(self):
        super().__init__()
    

    # Set mission
    def setMissionStatus(self, activeWaypoint=-1, missionStatus="Standby"):
        self.setData({
            "ACTIVE_WPT": activeWaypoint,
            "STATUS": missionStatus
        })


    # Set mission waypoints
    def setWaypoints(self, waypoints=[]):
        self.setData({
            "WAYPOINTS": waypoints
        })


# Developed by Aptus Engineering, Inc. <https://aptus.aero>
# See LICENSE.md file in project root directory

'''
Video chunk packet. Similar to image chunk, but meant to contain MPEG-encoded binary data.
'''

import struct
from datetime import datetime
from .packet import Packet, PacketType

class VideoPacket(Packet):
    '''
    Video packet
    '''

    pType = PacketType.VIDEO_CHUNK

    # Empty Constructor
    def __init__(self):
        super().__init__()

    
    # Pack chunk into packet
    # Chunk must be a segment of MPEG-encoded video
    # Maybe set up a UDP/TCP port for ffmpeg to stream video to and translate to websocket from here?
    @staticmethod
    def constructFromChunk(chunk, feedName='PRIMARY'):
        
        p = VideoPacket()

        # First 8 bytes represent UTC UNIX timestamp as 64-bit float (double-precision)
        utc = datetime.utcnow().timestamp()

        # Construct feedname
        feedNameBytes = bytes(''.join([ feedName[i] if i < len(feedName) else chr(0) for i in range(12) ]), 'utf-8')

        p.pData = bytes(struct.pack("d", utc)) + feedNameBytes + chunk

        return p
    

    # Get timestamp from packet
    def getTimestamp(self):
        utc = struct.unpack("d", self.pData[:8])[0]
        return datetime.fromtimestamp(utc)

    # Get image/video feed name
    def getFeedName(self):
        return ''.join([ chr(c) for c in self.pData[8:20] if c > 0 ])
    
    # Get chunk from packet
    def getChunk(self):
        return self.pData[20:]



import setuptools

with open("README.md", "r") as handle:
    long_description = handle.read()

setuptools.setup(
    name = "imperix",
    version = "0.2.3",
    description = "Imperix Node SDK including the NodeLink communication handlers that allow robots and drones to communicate with the Imperix streamers and Commander API.",
    long_description=long_description,
    long_description_content_type = "text/markdown",
    author = "Aptus Engineering Inc.",
    author_email = "software@aptusai.com",
    url = "https://bitbucket.org/pinetree-ai/imperix-node-sdk-python3",
    packages=setuptools.find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires = '>=3.6',
    install_requires=[
        'asyncio',
        'numpy',
        'Pillow',
        'requests',
        'websockets'
    ]
)